@@ -3,7 +3,7 @@ package Crypt::JWT;
 use strict;
 use warnings;
 
-our $VERSION = '0.034';
+our $VERSION = '0.037';
 
 use Exporter 'import';
 our %EXPORT_TAGS = ( all => [qw(decode_jwt encode_jwt)] );
@@ -163,6 +163,36 @@ sub _add_claims {
   $payload->{nbf} = $now + $args{relative_nbf} if defined $args{relative_nbf};
 }
 
+sub _verify_header {
+  my ($header, %args) = @_;
+
+  # currently we only check "typ" header parameter
+  my $check = $args{verify_typ};
+  return if !defined $check;
+
+  if (exists $header->{typ}) {
+    if (ref $check eq 'Regexp') {
+      my $value = $header->{typ};
+      $value = "" if !defined $value;
+      croak "JWT: typ header re check failed" unless $value =~ $check;
+    }
+    elsif (ref $check eq 'CODE') {
+      croak "JWT: typ header code check failed" unless $check->($header->{typ});
+    }
+    elsif (!ref $check) {
+      my $value = $header->{typ};
+      croak "JWT: typ header scalar check failed" unless defined $value && $value eq $check;
+    }
+    else {
+      croak "JWT: verify_typ must be Regexp, Scalar or CODE";
+    }
+  }
+  else {
+    croak "JWT: typ header required but missing"
+  }
+
+}
+
 sub _verify_claims {
   my ($payload, %args) = @_;
 
@@ -214,8 +244,34 @@ sub _verify_claims {
     }
   }
 
-  ### iss, sub, aud, jti
-  foreach my $claim (qw(iss sub aud jti)) {
+  ### aud
+  if (defined $args{verify_aud}) {
+    my $check = $args{verify_aud};
+    if (exists $payload->{aud}) {
+      my $match = 0;
+      # aud claim is a bit special as it can be either a string or an array of strings
+      my @aud_list = ref $payload->{aud} eq 'ARRAY' ? @{$payload->{aud}} : ( $payload->{aud} );
+      for my $value (@aud_list) {
+        if (ref $check eq 'Regexp') {
+          $value = "" if !defined $value;
+          $match = 1 if $value =~ $check;
+        }
+        elsif (ref $check eq 'CODE') {
+          $match = 1 if $check->($value);
+        }
+        elsif (!ref $check) {
+          $match = 1 if defined $value && $value eq $check;
+        }
+      }
+      croak "JWT: aud claim check failed" if !$match;
+    }
+    else {
+      croak "JWT: aud claim required but missing"
+    }
+  }
+
+  ### iss, sub, jti
+  foreach my $claim (qw(iss sub jti)) {
     my $check = $args{"verify_$claim"};
     next unless (defined $check);
 
@@ -539,6 +595,7 @@ sub _decode_jwe {
   $payload = _payload_unzip($payload, $header->{zip}) if $header->{zip};
   $payload = _payload_dec($payload, $args{decode_payload});
   _verify_claims($payload, %args); # croaks on error
+  _verify_header($header, %args); # croaks on error
   return ($header, $payload);
 }
 
@@ -696,6 +753,7 @@ sub _decode_jws {
   $payload = _payload_dec($payload, $args{decode_payload});
   _verify_claims($payload, %args); # croaks on error
   $header = { %$unprotected_header, %$header }; # merge headers
+  _verify_header($header, %args); # croaks on error
   return ($header, $payload);
 }
 
@@ -761,13 +819,18 @@ sub decode_jwt {
   if (!$args{token}) {
     croak "JWT: missing token";
   }
-  elsif ($args{token} =~ /^([a-zA-Z0-9_-]+)=*\.([a-zA-Z0-9_-]*)=*\.([a-zA-Z0-9_-]+)=*\.([a-zA-Z0-9_-]+)=*\.([a-zA-Z0-9_-]+)=*$/) {
-    # JWE token (5 segments)
-    ($header, $payload) = _decode_jwe($1, $2, $3, $4, $5, undef, {}, {}, %args);
-  }
-  elsif ($args{token} =~ /^([a-zA-Z0-9_-]+)=*\.([a-zA-Z0-9_-]+)=*\.([a-zA-Z0-9_-]*)=*$/) {
-    # JWS token (3 segments)
-    ($header, $payload) = _decode_jws($1, $2, $3, {}, %args);
+  my $token_re = $args{tolerate_padding}
+    ? qr/^([a-zA-Z0-9_-]+=*)\.([a-zA-Z0-9_-]*=*)\.([a-zA-Z0-9_-]*=*)(?:\.([a-zA-Z0-9_-]+=*)\.([a-zA-Z0-9_-]+=*))?$/
+    : qr/^([a-zA-Z0-9_-]+)=*\.([a-zA-Z0-9_-]*)=*\.([a-zA-Z0-9_-]*)=*(?:\.([a-zA-Z0-9_-]+)=*\.([a-zA-Z0-9_-]+)=*)?$/;
+  if ($args{token} =~ $token_re) {
+    if (defined($5) && length($5) > 0) {
+        # JWE token (5 segments)
+        ($header, $payload) = _decode_jwe($1, $2, $3, $4, $5, undef, {}, {}, %args);
+    }
+    else {
+        # JWS token (3 segments)
+        ($header, $payload) = _decode_jws($1, $2, $3, {}, %args);
+    }
   }
   elsif ($args{token} =~ /^\s*\{.*?\}\s*$/s) {
     my $hash = decode_json($args{token});
@@ -1022,6 +1085,17 @@ This parametes can be either a JWK Set JSON string (see RFC7517) or a perl HASH
   };
   my $payload = decode_jwt(token=>$t, kid_keys=>$keylist);
 
+You can use L<Crypt::PK::RSA/"export_key_jwk"> to generate a JWK for RSA:
+
+  my $pubkey = Crypt::PK::RSA->new('rs256-4096-public.pem');
+  my $jwk_hash = $pubkey->export_key_jwk('public', 1);
+  $jwk_hash->{kid} = 'key1';
+  my $keylist = {
+    keys => [
+      $jwk_hash,
+    ]
+  };
+
 The structure described above is used e.g. by L<https://www.googleapis.com/oauth2/v2/certs>
 
   use Mojo::UserAgent;
@@ -1036,7 +1110,7 @@ B<SINCE 0.019> we also support alternative structure used e.g. by L<https://www.
   my $payload = decode_jwt(token => $t, kid_keys => $google_certs);
 
 When the token header contains C<kid> item the corresponding key is looked up in C<kid_keys> list and used for token
-decoding (you do not need to pass the explicit key via C<key> parameter).
+decoding (you do not need to pass the explicit key via C<key> parameter). Add a kid header using L</"extra_headers">.
 
 B<INCOMPATIBLE CHANGE in 0.023:> When C<kid_keys> is specified it croaks if token header does not contain C<kid> value or
 if C<kid> was not found in C<kid_keys>.
@@ -1139,6 +1213,9 @@ C<Scalar> - 'aud' claim value has to be equal to given string (since 0.029)
 
 C<undef> (default) - do not verify 'aud' claim
 
+B<SINCE 0.036> we handle 'aud' claim when it contains an array of strings. In this case, the check should succeed if at least one
+value from the array matches. All checks (CODE, Regexp, Scalar) are performed individually against each member of the array of strings.
+
 =item verify_sub
 
 B<INCOMPATIBLE CHANGE in 0.024:> If C<verify_sub> is specified and
@@ -1199,6 +1276,24 @@ C<1> - do not check claims (iat, exp, nbf, iss, aud, sub, jti), B<BEWARE: DANGER
 
 C<0> (default) - check claims
 
+=item verify_typ
+
+B<SINCE 0.036>
+
+C<CODE ref> - subroutine (with 'typ' header parameter value passed as argument) should return C<true> otherwise verification fails
+
+C<Regexp ref> - 'typ' header parameter value has to match given regexp otherwise verification fails
+
+C<Scalar> - 'typ' header parameter value has to be equal to given string
+
+C<undef> (default) - do not verify 'typ' header parameter
+
+=item tolerate_padding
+
+C<0> (default) - ignore Base64 padding characters when validating signature
+
+C<1> - take account of Base64 padding characters when validating signature
+
 =back
 
 =head2 encode_jwt
@@ -1346,6 +1441,10 @@ iteration count (p2c) via C<extra_headers> like this:
  my $token = encode_jwt(payload=>$p, key=>$k, alg=>'PBES2-HS512+A256KW', extra_headers=>{p2c=8000, p2s=>32});
  #NOTE: handling of p2s header is a special case, in the end it is replaced with the generated salt
 
+You can also use this to specify a kid value (see L</"kid_keys">)
+
+ my $token = encode_jwt(payload=>$p, key=>$k, alg => 'RS256', extra_headers=>{kid=>'key1'});
+
 =item unprotected_headers
 
 A hash with additional integrity unprotected headers - JWS and JWE (not available for C<compact> serialization);
@@ -1404,4 +1503,4 @@ This program is free software; you can redistribute it and/or modify it under th
 
 =head1 COPYRIGHT
 
-Copyright (c) 2015-2021 DCIT, a.s. L<https://www.dcit.cz> / Karel Miko
+Copyright (c) 2015-2025 DCIT, a.s. L<https://www.dcit.cz> / Karel Miko
